import logging
from threading import Lock
from urlparse import urlparse, urlunparse, ParseResult

from miproxy.proxy import (
    ProxyHandler, RequestInterceptorPlugin,
    ResponseInterceptorPlugin, UnsupportedSchemeException)
from ssl import wrap_socket

log = logging.getLogger(__name__)


class TorProxyHandler(ProxyHandler):
    def __init__(self, tor_instance, *args, **kwargs):
        self.tor_instance = tor_instance
        ProxyHandler.__init__(self, *args, **kwargs)

    def _connect_to_host(self):
        # Get hostname and port to connect to
        if self.is_connect:
            self.hostname, self.port = self.path.split(':')
            log.debug('Connecting to %s:%s' % (self.hostname, self.port))
        else:
            u = urlparse(self.path)
            if u.scheme != 'http':
                raise UnsupportedSchemeException('Unknown scheme %s'
                                                 % repr(u.scheme))
            log.debug('Using %s to fetch %s'
                      % (self.tor_instance.name, self.path))
            self.hostname = u.hostname
            self.port = u.port or 80
            self.path = urlunparse(
                ParseResult(scheme='', netloc='', params=u.params,
                            path=u.path or '/', query=u.query,
                            fragment=u.fragment))

        # Connect to destination
        self._proxy_sock = None
        while self._proxy_sock is None:
            self._proxy_sock = self.tor_instance.create_socket(
                suppress_errors=True)
        self._proxy_sock.settimeout(10)
        self._proxy_sock.connect((self.hostname, int(self.port)))

        # Wrap socket if SSL is required
        if self.is_connect:
            self._proxy_sock = wrap_socket(self._proxy_sock)

    def mitm_request(self, data):
        # Register start time
        return ProxyHandler.mitm_request(self, data)

    def mitm_response(self, data):
        # Register end time and count requests
        # Keep stats and restart tor instance as needed
        return ProxyHandler.mitm_response(self, data)


class DebugInterceptor(RequestInterceptorPlugin, ResponseInterceptorPlugin):
    def do_request(self, data):
        print '>> %s' % repr(data[:100])
        return data

    def do_response(self, data):
        print '<< %s' % repr(data[:100])
        return data


def tor_proxy_handler_factory(tor_swarm):
    """ Return a factory for TorProxyHandlers bound to Tor instances. """
    tor_instances = tor_swarm.instances()
    generator_lock = Lock()  # Synchronize thread access to the generator.

    def factory(*args, **kwargs):
        while True:
            with generator_lock:
                tor_instance = next(tor_instances)
            if tor_instance.connected:
                break
        return TorProxyHandler(tor_instance, *args, **kwargs)

    return factory



from datetime import datetime
from itertools import chain, cycle
from os import path
from threading import Event, Lock, Thread
from time import sleep

import socks
from desub import desub

from proctor.socket import InstrumentedSocket

import logging
log = logging.getLogger(__name__)


class TorProcess(Thread):
    """ Runs and manages a Tor process in a thread.

    This class takes care of starting and stopping a Tor process, as well as
    monitoring connection times and the error rate and restarting the process
    when unhealthy.

    """
    def __init__(self, name, socks_port, control_port, base_work_dir,
                 boot_time_max=30, errors_max=10, conn_time_avg_max=2,
                 grace_time=30, sockets_max=None, resurrections_max=10):
        super(TorProcess, self).__init__()
        self.name = name
        self.socks_port = socks_port
        self.control_port = control_port
        self.base_work_dir = base_work_dir
        self.boot_time_max = boot_time_max
        self.errors_max = errors_max
        self.conn_time_avg_max = conn_time_avg_max
        self.grace_time = grace_time
        self.sockets_max = sockets_max
        self.resurrections_max = resurrections_max
        self._connected = Event()
        self._exclusive_access = Lock()
        self._ref_count = 0
        self._ref_count_lock = Lock()
        self._socket_count = 0
        self._socket_count_lock = Lock()
        self._stats_lock = Lock()
        self._stats_window = 200
        self._stoprequest = Event()
        self._terminated = False

    def run(self):
        """ Run and supervise the Tor process. """
        args = dict(CookieAuthentication=0, HashedControlPassword='',
                    ControlPort=self.control_port, PidFile=self.pid_file,
                    SocksPort=self.socks_port, DataDirectory=self.work_dir)
        args = map(str, chain(*(('--' + k, v) for k, v in args.iteritems())))
        tor = desub.join(['tor'] + args)
        self._start(tor)
        resurrections = 0
        while not self._stoprequest.is_set():
            if not tor.is_running():
                if resurrections >= self.resurrections_max:
                    log.error('Resurrected %s %s times, giving up.'
                              % (self.name, resurrections))
                    self._terminated = True
                    break
                resurrections += 1
                self._restart(tor, died=True)
            else:
                log.info('Started %s' % self.name)
            self.monitor(tor)

    def monitor(self, tor):
        """ Make sure Tor starts and stops when appropriate. """
        while tor.is_running():
            # Stop nicely when asked nicely.
            if self._stoprequest.wait(1):
                tor.stop()
                log.debug('Stopped %s' % self.name)
            # Check health and restart when appropriate.
            elif self._connected.is_set():
                errors, timing_avg, samples = self.get_stats()
                too_many_errors = errors > self.errors_max
                too_slow = timing_avg > self.conn_time_avg_max
                max_use_reached = (self.sockets_max
                                   and self._socket_count >= self.sockets_max)
                needs_restart = too_many_errors or too_slow or max_use_reached
                if self.age > self.grace_time and needs_restart:
                    self._restart(tor)
            else:
                out = tor.stdout.read()
                # Check for successful connection.
                if 'Bootstrapped 100%: Done.' in out:
                    self._connected.set()
                    log.info('%s is connected' % self.name)
                    self._start_time = datetime.utcnow()
                else:
                    # Check if initialization takes too long.
                    if self.time_since_boot > self.boot_time_max:
                        self._restart(tor, failed_boot=True)
                    # Check for socket binding failures.
                    else:
                        for port in [self.socks_port, self.control_port]:
                            if 'Could not bind to 127.0.0.1:%s' % port in out:
                                error = ('Could not bind %s to 127.0.0.1:%s'
                                        % (self.name, port))
                                log.warn(error)
                                self._terminated = True
                                break

    def stop(self):
        """ Signal the thread to stop itself. """
        self._stoprequest.set()

    @property
    def work_dir(self):
        return path.join(self.base_work_dir, self.name)

    @property
    def pid_file(self):
        return path.join(self.work_dir, 'pid')

    @property
    def connected(self):
        return self._connected.is_set()

    @property
    def age(self):
        """ Return the number of seconds since the Tor circuit is usable. """
        return (datetime.utcnow() - self._start_time).total_seconds()

    @property
    def terminated(self):
        return self._terminated

    @property
    def time_since_boot(self):
        """ Return the number of seconds since the last Tor process start. """
        return (datetime.utcnow() - self._boot_time).total_seconds()

    def _start(self, tor):
        """ Start a Tor process. """
        with self._stats_lock:
            self._boot_time = datetime.utcnow()
            self._socket_count = 0
            self._stats_errors = list()
            self._stats_timing = list()
        tor.start()

    def _restart(self, tor, failed_boot=False, died=False):
        """ Safely replace a Tor instance with a fresh one. """
        with self._exclusive_access:  # Prevent creating sockets.
            # Wait until all sockets have finished.
            wait_start = datetime.utcnow()
            while self._ref_count > 0:
                if (datetime.utcnow() - wait_start).total_seconds() > 30:
                    log.error('Likely got a ref_count accounting error in %s'
                              % self.name)
                    self._ref_count = 0
                    break
                sleep(1)
            self._connected.clear()
            if failed_boot:
                log.warn('Restarting %s (did not initialize in time)'
                         % self.name)
            elif died:
                log.warn('Resurrected %s' % self.name)
            else:
                errors, timing_avg, samples = self.get_stats()
                log.warn(('Restarting %s '
                          '(errors: %s, avg time: %s, count: %s, age: %s)')
                         % (self.name, errors, timing_avg, self._socket_count,
                            int(self.age)))
            tor.stop()
            self._start(tor)

    def _inc_socket_count(self):
        """ Increment the internal socket counter. """
        with self._socket_count_lock:
            self._socket_count += 1

    def _inc_ref_count(self):
        """ Increment the internal reference counter. """
        with self._ref_count_lock:
            self._ref_count += 1

    def _dec_ref_count(self):
        """ Decrement the internal reference counter. """
        with self._ref_count_lock:
            self._ref_count -= 1

    def _receive_stats(self, timing, errors):
        """ Maintain connection statistics over time. """
        with self._stats_lock:
            self._stats_errors.append(errors)
            self._stats_timing.append(timing)
            if len(self._stats_errors) > self._stats_window:
                self._stats_errors = self._stats_errors[-self._stats_window:]
                self._stats_timing = self._stats_timing[-self._stats_window:]
            # We consider the socket at end of life when it sends the stats.
            self._dec_ref_count()

    def get_stats(self):
        """ Return current statistics. """
        with self._stats_lock:
            samples = len(self._stats_timing)
            errors = sum(self._stats_errors)
            timing_avg = sum(self._stats_timing) / (samples or 1)
            return errors, timing_avg, samples

    def create_socket(self, suppress_errors=False, *args, **kwargs):
        """ Return an InstrumentedSocket that will connect through Tor. """
        if self.connected:
            if not self._exclusive_access.acquire(False):
                return None
            try:
                sock = InstrumentedSocket(self._receive_stats, *args, **kwargs)
                args = (socks.PROXY_TYPE_SOCKS4, 'localhost', self.socks_port,
                        True, None, None)  # rdns, username, password
                sock.setproxy(*args)
                # Keep track of how many sockets are using this Tor instance.
                self._inc_ref_count()
                self._inc_socket_count()
                return sock
            finally:
                self._exclusive_access.release()
        elif suppress_errors:
            sleep(0.1)  # Prevent fast spinning in (the proxy code) caused by
                        # a race condition when Tor restarts.
            return None
        else:
            raise RuntimeError('%s not yet connected.' % self.name)


class TorSwarm(object):
    """ Manages a number of Tor processes. """
    def __init__(self, base_socks_port, base_control_port, work_dir,
                 sockets_max, **kwargs):
        self.base_socks_port = base_socks_port
        self.base_control_port = base_control_port
        self.work_dir = work_dir
        self.sockets_max = sockets_max
        self.kwargs = kwargs
        self._instances = list()

    def instances(self):
        """ Return an infinite generator cycling through Tor instances. """
        for instance in cycle(self._instances):
            if instance.terminated:
                alive = list(i for i in self._instances if not i.terminated)
                if len(alive) == 0:
                    log.critical('No alive Tor instance left. Bailing out.')
                    return
            yield instance

    def start(self, num_instances):
        """ Start and return the Tor processes. """
        log.info('Starting Tor swarm with %d instances...' % num_instances)
        self._instances = list()
        for i in range(num_instances):
            tor = TorProcess('tor-%d' % i, self.base_socks_port + i,
                            self.base_control_port + i, self.work_dir,
                            sockets_max=self.sockets_max, **self.kwargs)
            self._instances.append(tor)
            tor.start()
            sleep(0.1)
        return self._instances

    def stop(self):
        """ Stop the Tor processes and wait for their completion. """
        for tor in self._instances:
            tor.stop()
            tor.join()



""" An HTTP proxy that routes requests through a number of Tor circuits. """

import logging
import sys
from argparse import ArgumentParser
from shutil import rmtree
from tempfile import mkdtemp
from time import sleep

from miproxy.proxy import AsyncMitmProxy

from proctor.vendor.exit import handle_exit

LOG_FORMAT = '%(asctime)s,%(msecs)03d %(levelname)-5.5s [%(name)s] %(message)s'


def get_args_parser():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('-d', '--work-dir', help='Working directory')
    parser.add_argument('-p', '--port', type=int, default=8080,
                        help='Proxy server listening port')
    parser.add_argument('-s', '--base-socks-port', type=int, default=19050,
                        help='Base socks port for the Tor processes')
    parser.add_argument('-c', '--base-control-port', type=int, default=18118,
                        help='Base control port for the Tor processes')
    parser.add_argument('-n', '--instances', type=int, default=2,
                        help='Number of Tor processes to launch')
    parser.add_argument('-m', '--max-use', type=int,
                        help='Max number of requests before replacing '
                             'Tor processes')
    parser.add_argument('-t', '--max-conn-time', type=float, default=2,
                        help='Number of Tor processes to launch')
    return parser


def parse_args():
    parser = get_args_parser()
    parser.add_argument('-l', '--loglevel', default='INFO',
                        choices=('CRITICAL', 'ERROR', 'WARN', 'INFO', 'DEBUG'),
                        help='Display messages above this log level')
    return parser.parse_args()


def run_proxy(port, base_socks_port, base_control_port, work_dir,
              num_instances, sockets_max, **kwargs):
    # Imported here so that the logging module could be initialized by another
    # script that would import from the present module. Not sure that's the
    # best way to accomplish this though.
    from .tor import TorSwarm
    from .proxy import tor_proxy_handler_factory

    log = logging.getLogger(__name__)

    proxy = None
    tor_swarm = None

    def kill_handler():
        log.warn('Interrupted, stopping server')
        try:
            if proxy:
                proxy.server_close()
        finally:
            if tor_swarm is not None:
                tor_swarm.stop()

    with handle_exit(kill_handler):
        tor_swarm = TorSwarm(base_socks_port, base_control_port, work_dir,
                             sockets_max, **kwargs)
        tor_instances = tor_swarm.start(num_instances)
        log.debug('Waiting for at least one connected Tor instance...')
        while not [t for t in tor_instances if t.connected]:
            if len(list(i for i in tor_instances if not i.terminated)) == 0:
                log.critical('No alive Tor instance left. Bailing out.')
                sys.exit(1)
            sleep(0.25)
        handler_factory = tor_proxy_handler_factory(tor_swarm)
        proxy = AsyncMitmProxy(server_address=('', port),
                               RequestHandlerClass=handler_factory)
        log.info('Starting proxy server on port %s' % port)
        proxy.serve_forever()


def main():
    args = parse_args()
    work_dir = args.work_dir or mkdtemp()
    logging.basicConfig(level=getattr(logging, args.loglevel),
                        format=LOG_FORMAT)
    try:
        run_proxy(args.port, args.base_socks_port, args.base_control_port,
                  work_dir, args.instances, args.max_use,
                  conn_time_avg_max=args.max_conn_time)
    finally:
        if not args.work_dir:
            rmtree(work_dir)


if __name__ == '__main__':
    main()



# encoding: utf-8

from datetime import date

__author__ = 'Nicolas Cadou'
__copyright__ = 'Copyright 2013-%d, Nicolas Cadou' % date.today().year
__credits__ = ['Nadeem Douba', 'Dan-Haim']

__license__ = 'New BSD'
__version__ = '0.1'
__maintainer__ = 'Nicolas Cadou'
__email__ = 'ncadou@cadou.ca'
__status__ = 'Development'
__url__ = 'http://pypi.python.org/pypi/onion-proctor'

__all__ = ['proxy', 'scripts', 'socket', 'tor']



# Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
# License: MIT
#
# URL: http://code.activestate.com/recipes/577997-handle-exit-context-manager/

from __future__ import with_statement
import contextlib
import signal
import sys


def _sigterm_handler(signum, frame):
    sys.exit(0)
_sigterm_handler.__enter_ctx__ = False


@contextlib.contextmanager
def handle_exit(callback=None, append=False):
    """A context manager which properly handles SIGTERM and SIGINT
    (KeyboardInterrupt) signals, registering a function which is
    guaranteed to be called after signals are received.
    Also, it makes sure to execute previously registered signal
    handlers as well (if any).

    >>> app = App()
    >>> with handle_exit(app.stop):
    ...     app.start()
    ...
    >>>

    If append == False raise RuntimeError if there's already a handler
    registered for SIGTERM, otherwise both new and old handlers are
    executed in this order.
    """
    old_handler = signal.signal(signal.SIGTERM, _sigterm_handler)
    if (old_handler != signal.SIG_DFL) and (old_handler != _sigterm_handler):
        if not append:
            raise RuntimeError("there is already a handler registered for "
                               "SIGTERM: %r" % old_handler)

        def handler(signum, frame):
            try:
                _sigterm_handler(signum, frame)
            finally:
                old_handler(signum, frame)
        signal.signal(signal.SIGTERM, handler)

    if _sigterm_handler.__enter_ctx__:
        raise RuntimeError("can't use nested contexts")
    _sigterm_handler.__enter_ctx__ = True

    try:
        yield
    except KeyboardInterrupt:
        pass
    except SystemExit, err:
        # code != 0 refers to an application error (e.g. explicit
        # sys.exit('some error') call).
        # We don't want that to pass silently.
        # Nevertheless, the 'finally' clause below will always
        # be executed.
        if err.code != 0:
            raise
    finally:
        _sigterm_handler.__enter_ctx__ = False
        if callback is not None:
            callback()


if __name__ == '__main__':
    # ===============================================================
    # --- test suite
    # ===============================================================

    import unittest
    import os

    class TestOnExit(unittest.TestCase):

        def setUp(self):
            # reset signal handlers
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            self.flag = None

        def tearDown(self):
            # make sure we exited the ctx manager
            self.assertTrue(self.flag is not None)

        def test_base(self):
            with handle_exit():
                pass
            self.flag = True

        def test_callback(self):
            callback = []
            with handle_exit(lambda: callback.append(None)):
                pass
            self.flag = True
            self.assertEqual(callback, [None])

        def test_kinterrupt(self):
            with handle_exit():
                raise KeyboardInterrupt
            self.flag = True

        def test_sigterm(self):
            with handle_exit():
                os.kill(os.getpid(), signal.SIGTERM)
            self.flag = True

        def test_sigint(self):
            with handle_exit():
                os.kill(os.getpid(), signal.SIGINT)
            self.flag = True

        def test_sigterm_old(self):
            # make sure the old handler gets executed
            queue = []
            signal.signal(signal.SIGTERM, lambda s, f: queue.append('old'))
            with handle_exit(lambda: queue.append('new'), append=True):
                os.kill(os.getpid(), signal.SIGTERM)
            self.flag = True
            self.assertEqual(queue, ['old', 'new'])

        def test_sigint_old(self):
            # make sure the old handler gets executed
            queue = []
            signal.signal(signal.SIGINT, lambda s, f: queue.append('old'))
            with handle_exit(lambda: queue.append('new'), append=True):
                os.kill(os.getpid(), signal.SIGINT)
            self.flag = True
            self.assertEqual(queue, ['old', 'new'])

        def test_no_append(self):
            # make sure we can't use the context manager if there's
            # already a handler registered for SIGTERM
            signal.signal(signal.SIGTERM, lambda s, f: sys.exit(0))
            try:
                with handle_exit(lambda: flag.append(None)):
                    pass
            except RuntimeError:
                pass
            else:
                self.fail("exception not raised")
            finally:
                self.flag = True

        def test_nested_context(self):
            self.flag = True
            try:
                with handle_exit():
                    with handle_exit():
                        pass
            except RuntimeError:
                pass
            else:
                self.fail("exception not raised")

    unittest.main()



from contextlib import contextmanager
from datetime import datetime

import socks


class InstrumentedSocket(socks.socksocket):
    """ A socket that maintains timing info about connection/disconnection.

    The timing info will be sent back once to the callback on either socket
    shutdown(), close(), or on any error.

    """
    def __init__(self, callback, *args, **kwargs):
        self._callback = callback
        self._called_back = False
        self._error_count = 0
        self._total_time = 0
        socks.socksocket.__init__(self, *args, **kwargs)

    @contextmanager
    def _timer(self):
        """ Context manager that measures time spent and count errors. """
        def update_timing():
            self._total_time += (datetime.now() - start_time).total_seconds()

        start_time = datetime.now()
        try:
            try:
                yield
            finally:
                update_timing()
        except:
            self._error_count += 1
            self._do_callback()
            raise

    @contextmanager
    def _callback_on_error(self):
        """ Context manager that sends stats to the callback on errors. """
        try:
            yield
        except:
            self._error_count += 1
            self._do_callback()
            raise

    def _do_callback(self):
        """ Communicate back socket connection statistics. """
        if not self._called_back:
            self._callback(self._total_time, self._error_count)
            self._called_back = True

    def connect(self, address):
        with self._timer():
            return socks.socksocket.connect(self, address)

    def connect_ex(self, address):
        with self._timer():
            return socks.socksocket.connect_ex(self, address)

    def send(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.send(self, *args, **kwargs)

    def sendall(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.sendall(self, *args, **kwargs)

    def sendto(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.sendto(self, *args, **kwargs)

    def sendblocking(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.sendblocking(self, *args, **kwargs)

    def recv(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.recv(self, *args, **kwargs)

    def recvfrom(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.recvfrom(self, *args, **kwargs)

    def recvfrom_into(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.recvfrom_into(self, *args, **kwargs)

    def recv_into(self, *args, **kwargs):
        with self._callback_on_error():
            return socks.socksocket.recv_into(self, *args, **kwargs)

    def shutdown(self, how):
        with self._timer():
            result = socks.socksocket.shutdown(self, how)
        self._do_callback()
        return result

    def close(self):
        with self._timer():
            result = socks.socksocket.close(self)
        self._do_callback()
        return result



# encoding: utf-8

import os

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

import proctor


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='onion-proctor',
      version=proctor.__version__,
      author=proctor.__author__,
      author_email=proctor.__email__,
      maintainer=proctor.__maintainer__,
      maintainer_email=proctor.__email__,
      url=proctor.__url__,
      long_description=read('README.md'),
      license='LICENSE.txt',
      packages=find_packages(),
      include_package_data=True,
      install_requires=['desub', 'pymiproxy', 'SocksiPy-branch'],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Plugins',
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Topic :: Documentation',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Internet :: WWW/HTTP'],
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'proctor = proctor.scripts:main']})
